/******************************************************************************************

   Hex7.java - 7x7 Hex board game
   Version 2.0
   www.mazeworks.com
   
   Copyright (c) 2002 Robert Kirkland
   All Rights Reserved.

   Permission is hereby granted, free of charge, to any person obtaining a copy 
   of this software and associated documentation files (the "Software"), to
   deal in the Software without restriction, including without limitation the
   rights to use, copy, modify, merge, publish, distribute, and/or sell copies
   of the Software, and to permit persons to whom the Software is furnished to
   do so, provided that the above copyright notice and this permission notice
   appear in all copies of the Software and that both the above copyright
   notice and this permission notice appear in supporting documentation.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER INCLUDED IN THIS NOTICE BE LIABLE FOR
   ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Except as contained in this notice, the name of the copyright holder shall
   not be used in advertising or otherwise to promote the sale, use or other
   dealings in this Software without prior written authorization of the
   copyright holder.

******************************************************************************************/

import java.applet.* ;
import java.awt.* ;
import java.awt.image.*;
import java.net.* ;
import java.util.* ;
	
/******** HEX7 ********
   main class: initializes applet panel, starts games,
   runs AI move thread, handles player moves and UI events
*/   
public class Hex7 extends Applet implements Runnable {
   // applet info
   public String getAppletInfo() { return "Hex 2.0 - Copyright (c) 2002 Robert Kirkland"; }
   String parameterInfo[][] = {
      {"bgColor",   "integer", "Background color (24-bit RGB hex int)"},
      {"title",     "string",  "Applet title text"},
      {"newGame",   "string",  "New Game button text"},
      {"white",     "string",  "Player Side White text"},
      {"black",     "string",  "Player Side Black text"},
      {"level",     "string",  "Level label text"},
      {"basic",     "string",  "Basic level text"},
      {"advanced",  "string",  "Advanced level text"},
      {"yourMove",  "string",  "Player Move status text"},
      {"thinking",  "string",  "Computer Thinking status text"},
      {"whiteWins", "string",  "White Wins status text"},
      {"blackWins", "string",  "Black Wins status text"} } ;
   public String[][] getParameterInfo() { return parameterInfo ; }

   static final Color  DF_BG_COLOR  =Color.white ;   // default background color
   static final String DF_sTITLE     ="HEX-7",       // default text strings
                       DF_sNEW_GAME  ="NEW GAME",
                       DF_sWHITE     ="White",  
                       DF_sBLACK     ="Black",
                       DF_sLEVEL     ="LEVEL",
                       DF_sBASIC     ="Basic",
                       DF_sADVANCED  ="Advanced",
                       DF_sYOUR_MOVE ="Your move..",
                       DF_sTHINKING  ="Thinking..",
                       DF_sWHITE_WINS="White wins !",
                       DF_sBLACK_WINS="Black wins !" ;
   static final int bcWidth=446, bcHeight=247, PLAYER=0, COMPUTER=1, WHITE=1, BLACK=2 ;
   protected Color BG_COLOR ;
   protected String sTITLE, sNEW_GAME, sWHITE, sBLACK, sLEVEL, sBASIC, sADVANCED,
                    sYOUR_MOVE, sTHINKING, sWHITE_WINS, sBLACK_WINS ;
   private boolean playerTurn, gameOver ;
   private int playerColor=WHITE, computerColor=BLACK, ply ;
   private StaticEval stat ;
   private Board bd ;
   private BoardCanvas bc ;
   private ControlPanel cp ;
   private Image boardImage, gifletImage ;
   private AI piet ;
   private Thread pietMoveThread ;
	
   public void init() {
      // --- get applet parameters ---
      // background color
      String bgColorString = getParameter("bgColor") ;
      try { BG_COLOR = bgColorString==null ? DF_BG_COLOR : new Color(Integer.parseInt(bgColorString,16)) ; } 
      catch (NumberFormatException e) {
         BG_COLOR = DF_BG_COLOR ;
         System.out.println("ERROR: Invalid bgColor parameter.") ;
      }
      // text
      sTITLE = getParameter("title") ; if (sTITLE==null) sTITLE = new String(DF_sTITLE) ;
      sNEW_GAME = getParameter("newGame") ; if (sNEW_GAME==null) sNEW_GAME = new String(DF_sNEW_GAME) ;
      sWHITE = getParameter("white") ; if (sWHITE==null) sWHITE = new String(DF_sWHITE) ;
      sBLACK = getParameter("black") ; if (sBLACK==null) sBLACK = new String(DF_sBLACK) ;
      sLEVEL = getParameter("level") ; if (sLEVEL==null) sLEVEL = new String(DF_sLEVEL) ;
      sBASIC = getParameter("basic") ; if (sBASIC==null) sBASIC = new String(DF_sBASIC) ;
      sADVANCED = getParameter("advanced") ; if (sADVANCED==null) sADVANCED = new String(DF_sADVANCED) ;
      sYOUR_MOVE = getParameter("yourMove") ; if (sYOUR_MOVE==null) sYOUR_MOVE = new String(DF_sYOUR_MOVE) ;
      sTHINKING = getParameter("thinking") ; if (sTHINKING==null) sTHINKING = new String(DF_sTHINKING) ;
      sWHITE_WINS = getParameter("whiteWins") ; if (sWHITE_WINS==null) sWHITE_WINS = new String(DF_sWHITE_WINS) ;
      sBLACK_WINS = getParameter("blackWins") ; if (sBLACK_WINS==null) sBLACK_WINS = new String(DF_sBLACK_WINS) ;

		// load image files
		MediaTracker tracker = new MediaTracker(this) ;
		URL url = getCodeBase() ;
		boardImage = getImage(url,"board7.gif") ;
      gifletImage = getImage(url,"giflets.gif") ;
		tracker.addImage(boardImage,0) ;
      tracker.addImage(gifletImage,1) ;
		try { tracker.waitForAll(); }
		catch (InterruptedException e) {}

      // initialize applet layout
      setBackground(BG_COLOR) ;
      setLayout(new FlowLayout(FlowLayout.CENTER,10,10)) ;
      // create board canvas
      bc = new BoardCanvas(boardImage,gifletImage,this) ;
      bc.resize(bcWidth,bcHeight) ;
      // create control panel
      cp = new ControlPanel(this) ;
      Panel p1 = new Panel() ;
      p1.setBackground(Color.lightGray) ;
      p1.setLayout(new FlowLayout(FlowLayout.CENTER,5,5)) ;
      Panel p2 = new Panel() ;
      p2.setBackground(Color.gray) ;
      p2.setLayout(new FlowLayout(FlowLayout.CENTER,2,2)) ;
      p1.add(cp) ;
      p2.add(p1) ;
      add(bc) ;
      add(p2) ;
      validate() ;
      // force class file load
      BestMove bm = new BestMove() ;
            
      newGame() ;
	}
	// start
   void newGame() {
      System.gc() ;
      stop() ;
      playerTurn = (playerColor==WHITE) ? true : false ;
      gameOver = false ; 
      ply = 1 ;
      bd = new Board(bc,this) ;
      stat = new StaticEval(bd,this) ;
      piet = new AI(bd,stat,cp,this) ;
      bc.drawBoard(bd) ;
      // white moves first
      if (!playerTurn) pietMoveStart() ;
      else cp.setStatus(sYOUR_MOVE) ;
      bc.requestFocus() ;
   }      
   private void endGame(int side) {
      if (side==COMPUTER) {
         // computer win
         if (computerColor==WHITE) cp.setStatus(sWHITE_WINS) ;
         else cp.setStatus(sBLACK_WINS) ;
      } 
      else {
         // player win
         if (playerColor==WHITE) cp.setStatus(sWHITE_WINS) ;
         else cp.setStatus(sBLACK_WINS) ;
      }
      gameOver = true ;
   }
   private void playerMove(int i, int j) {
      if (ply!=1||i!=3||j!=3) {
         if (bd.setHex(i,j,playerColor)) {
            bc.drawBoard(bd) ;
            if (bd.isWin(playerColor)) endGame(PLAYER) ;
            // followed by AI move if not a win for player
            else {
               ply++ ;
               pietMoveStart() ;
            }
         }
      }
   }
   // start AI move thread
   public void pietMoveStart() {
      if (pietMoveThread == null) {
         pietMoveThread = new Thread(this) ;
         pietMoveThread.start() ;
      }
   }
   // kill thread
   public void stop() {
      if (pietMoveThread != null) {
         pietMoveThread.stop() ;
         pietMoveThread = null ;
      }
   }
   // run thread
   public void run() { 
      playerTurn = false ;
      cp.setStatus(sTHINKING) ;
      piet.makeMove() ; 
      bc.drawBoard(bd) ;
      if (bd.isWin(computerColor)) endGame(COMPUTER) ;
      else {
         ply++ ;
         playerTurn = true ;
         cp.setStatus(sYOUR_MOVE) ;
      }
      pietMoveThread = null ;
   }
   void sleep(int ms) {
      try { Thread.currentThread().sleep(ms) ; } 
      catch (InterruptedException e) {}
   }
   int getColor(int side) { 
      return (side==PLAYER)? playerColor : computerColor ; 
   }
   int getOtherColor(int color) { 
      return (color==WHITE)? BLACK : WHITE ; 
   }
   int getPly() { return ply ; }

   // board canvas mouse event
   void selectHex(int i,int j) { if (!gameOver&&playerTurn) playerMove(i,j) ; }

   void setPlayerColor(int color) {
      if (gameOver||playerTurn) {
         if ((color==BLACK)&&(playerColor==WHITE)) {
            playerColor = BLACK ; computerColor = WHITE ;
         }
         else if ((color==WHITE)&&(playerColor==BLACK)) {
            playerColor = WHITE ; computerColor = BLACK ;
         }
         if (!gameOver) pietMoveStart() ;
      }
      bc.requestFocus() ;
   }
   void setLevel() { bc.requestFocus() ; }
}  
/******** BOARD ********
   controls board status, rules of the game,
   tests for win by either side
*/   
final class Board {
	static final int SIZE=7, EMPTY=0, WHITE=1, BLACK=2,
                    SW=1, SE=2, E=3, NE=4, NW=5, W=6,
                    SW_B=7, S_B=8, SE_B=9, NE_B=10, N_B=11, NW_B=12 ;
   private boolean visited[][] = new boolean[SIZE][SIZE] ;
	private int bd[][]=new int[SIZE][SIZE], stack[]=new int[25], stackPtr=0 ;
   private BoardCanvas bc ;
   private Hex7 main ;
	
   // constructor
   Board(BoardCanvas bc,Hex7 main) {
      this.bc = bc ;
      this.main = main ;
      for (int i=0; i<SIZE; i++)
         for (int j=0; j<SIZE; j++) bd[i][j] = 0 ;
   }
	boolean isEmpty(int i,int j) { return (bd[i][j] == EMPTY) ; }
	boolean isEmpty(Point h) { return (bd[h.x][h.y] == EMPTY) ; }
   boolean isWhite(int i,int j) { return (bd[i][j] == WHITE) ; }
   boolean isBlack(int i,int j) { return (bd[i][j] == BLACK) ; }
   boolean isColor(int i,int j,int color) { return (bd[i][j] == color) ; }
	int getHex(int i,int j) { return bd[i][j] ; }
   void setEmpty(int i,int j) { bd[i][j] = EMPTY ; }
   void setEmpty(Point h) { bd[h.x][h.y] = EMPTY ; }
	boolean setHex(int i,int j,int color) { 
		if ( (i>=0)&&(i<SIZE)&&(j>=0)&&(j<SIZE)&&(isEmpty(i,j)) ) { 
			bd[i][j] = color ;
			return true ;
		}
		else return false ;
	}
   boolean setHex(Point p,int color) { 
      return setHex(p.x,p.y,color) ;
   }
   // make the piece blink for computer move
   void computerMove(Point p,int color) {
      if (setHex(p,color)&&(main.getPly()!=1)) {
         bc.drawBoard(this) ;
         main.sleep(400) ;
         setEmpty(p) ;  
         bc.drawBoard(this) ;
         main.sleep(400) ;
         setHex(p,color) ;
      }
   }
   boolean isVisited(int i,int j) { return visited[i][j] ; }
   void setVisited(int i,int j,boolean v) { visited[i][j] = v ; }
   void clearVisited() {
      for (int i=0; i<SIZE; i++)
         for (int j=0; j<SIZE; j++) visited[i][j] = false ;
   }
   // main routine to check for win 
   boolean isWin(int color) {
      clearVisited() ;
      if (color==WHITE) return isWinWhite() ;
      else return isWinBlack() ;
   }
   private boolean isWinWhite() {
      int x=0, y=0 ;

      for (int i=0; i<SIZE; i++) {
         if ((bd[i][0]==WHITE)&&(!visited[i][0])) {
            x = i ; y = 0 ; stackPtr = 0 ;
            visited[x][y] = true ;
            if (isWinDFS(x,y,WHITE)) return true ;
         }
      }
      return false ;
   }
   private boolean isWinBlack() {
      int x=0, y=0 ;

      for (int j=0; j<SIZE; j++) {
         if ((bd[0][j]==BLACK)&&(!visited[0][j])) {
            x = 0 ; y = j ; stackPtr = 0 ;
            visited[x][y] = true ;
            if (isWinDFS(x,y,BLACK)) return true ;
         }
      }
      return false ;
   }
   private boolean isWinDFS(int x,int y,int color) {
      int direction=0 ;

      while (true) {
         // look for a link
         direction = scanLinks(x,y,color) ;
         if (direction!=0) {
            // found - make it current
            stack[stackPtr++] = direction ;
            switch (direction) {
               case SW: y++ ; break ;
               case SE: x++ ; y++ ; break ;
               case E : x++ ; break ;
               case NE: y-- ; break ;
               case NW: x-- ; y-- ; break ;
               case W : x-- ; break ;
            }
            // check for win
            if (color==WHITE) { if (y==SIZE-1) return true ; }
            else if (x==SIZE-1) return true ;
            visited[x][y] = true ;
         }
         else {
            // no path, try next start point
            if (stackPtr==0) return false ;
            // at end of path, back up
            direction = stack[--stackPtr] ;
            switch (direction) {
               case SW: y-- ; break ;
               case SE: x-- ; y-- ; break ;
               case E : x-- ; break ;
               case NE: y++ ; break ;
               case NW: x++ ; y++ ; break ;
               case W : x++ ; break ;
            }
         }
      }
   }
   int scanLinks(int x,int y,int color) {
      int dir=0 ;
      // search for unvisited links
      if ((isLinkSW(x,y,color))&&(!visited[x][y+1])) dir = SW ;
      else if ((isLinkSE(x,y,color))&&(!visited[x+1][y+1])) dir = SE ;
      else if ((isLinkE(x,y,color))&&(!visited[x+1][y])) dir = E ;
      else if ((isLinkNE(x,y,color))&&(!visited[x][y-1])) dir = NE ;
      else if ((isLinkNW(x,y,color))&&(!visited[x-1][y-1])) dir = NW ;
      else if ((isLinkW(x,y,color))&&(!visited[x-1][y])) dir = W ;
      return dir ;
   }
   // links
   boolean isLinkSW(int i,int j,int color) {
      if ((j<SIZE-1)&&(isColor(i,j+1,color))) return true ;
      else return false ;
   }
   boolean isLinkSE(int i,int j,int color) {
      if ((i<SIZE-1)&&(j<SIZE-1)&&(isColor(i+1,j+1,color))) return true ;
      else return false ;
   }
   boolean isLinkE(int i,int j,int color) {
      if ((i<SIZE-1)&&(isColor(i+1,j,color))) return true ;
      else return false ;
   }
   boolean isLinkNE(int i,int j,int color) {
      if ((j>0)&&(isColor(i,j-1,color)))  return true ;
      else return false ;
   }
   boolean isLinkNW(int i,int j,int color) {
      if ((i>0)&&(j>0)&&(isColor(i-1,j-1,color))) return true ;
      else return false ;
   }
   boolean isLinkW(int i,int j,int color) {
      if ((i>0)&&(isColor(i-1,j,color))) return true ;
      else return false ;
   }
   // Bridge methods are used by static evaluator
   int scanBridges(int x,int y,int color) {
      int dir=0 ;
      // search for unvisited bridges
      if ((isBridgeSW(x,y,color))&&(!visited[x-1][y+1])) dir = SW_B ;
      else if ((isBridgeS(x,y,color))&&(!visited[x+1][y+2])) dir = S_B ;
      else if ((isBridgeSE(x,y,color))&&(!visited[x+2][y+1])) dir = SE_B ;
      else if ((isBridgeNE(x,y,color))&&(!visited[x+1][y-1])) dir = NE_B ;
      else if ((isBridgeN(x,y,color))&&(!visited[x-1][y-2])) dir = N_B ;
      else if ((isBridgeNW(x,y,color))&&(!visited[x-2][y-1])) dir = NW_B ;
      return dir ;
   }
   // bridges
   boolean isBridgeSW(int i,int j,int color) {
      if ((i>0)&&(j<SIZE-1)&&(isColor(i-1,j+1,color))&&
         (isEmpty(i-1,j))&&(isEmpty(i,j+1))) return true ;
      else return false ;
   }
   boolean isBridgeS(int i,int j,int color) {
      if ((i<SIZE-1)&&(j<SIZE-2)&&(isColor(i+1,j+2,color))&&
         (isEmpty(i,j+1))&&(isEmpty(i+1,j+1))) return true ;
      else return false ;
   }
   boolean isBridgeSE(int i,int j,int color) {
      if ((i<SIZE-2)&&(j<SIZE-1)&&(isColor(i+2,j+1,color))&&
          (isEmpty(i+1,j))&&(isEmpty(i+1,j+1))) return true ;
      else return false ;
   }
   boolean isBridgeNE(int i,int j,int color) {
      if ((i<SIZE-1)&&(j>0)&&(isColor(i+1,j-1,color))&&
          (isEmpty(i+1,j))&&(isEmpty(i,j-1))) return true ;
      else return false ;
   }
   boolean isBridgeN(int i,int j,int color) {
      if ((i>0)&&(j>1)&&(isColor(i-1,j-2,color))&&
         (isEmpty(i,j-1))&&(isEmpty(i-1,j-1))) return true ;
      else return false ;
   }
   boolean isBridgeNW(int i,int j,int color) {
      if ((i>1)&&(j>0)&&(isColor(i-2,j-1,color))&&
          (isEmpty(i-1,j))&&(isEmpty(i-1,j-1))) return true ;
      else return false ;
   }
}			
/******** AI ********
   generates computer moves
*/   
final class AI {
   static final int PLAYER=0, COMPUTER=1, EMPTY=0, WHITE=1, BLACK=2, 
                    PLAYER_WIN=-999999, COMPUTER_WIN=999999 ;
   static final byte hexes[][] = {
         // cols 0 and 1 hold all hexes from center outward
         // cols 2 and 3 hold 2nd-ply response move 
         // inner diamond
         {3,3,0,0},{2,2,3,4},{4,4,3,2},{3,2,3,3},{4,3,4,4},
         {2,3,2,2},{3,4,3,3},{4,2,3,3},{2,4,3,3},
         // middle ring
         {1,1,2,3},{5,5,4,3},{2,1,3,3},{5,4,5,5},{1,2,1,1},{4,5,3,3},
         {3,1,3,2},{5,3,4,2},{1,3,2,4},{3,5,3,4},{4,1,4,2},
         {5,2,3,3},{1,4,3,3},{2,5,2,4},{5,1,5,5},{1,5,1,1},
         // outer ring
         {0,0,1,2},{6,6,5,4},{1,0,3,3},{6,5,3,3},{0,1,3,3},{5,6,3,3},
         {2,0,3,3},{6,4,5,5},{0,2,1,1},{4,6,3,3},{3,0,3,3},{6,3,3,3},
         {0,3,3,3},{3,6,3,3},{4,0,3,3},{6,2,3,3},{0,4,3,3},{2,6,3,3},
         {5,0,3,3},{6,1,3,3},{0,5,3,3},{1,6,3,3},{6,0,3,3},{0,6,3,3} } ;
   private int ply, maxDepth ;
   private BestMove best ;
   private Point move ;
   private Board bd ;
   private StaticEval stat ;
   private ControlPanel cp ;
   private Hex7 main ;

   // constructor
   AI(Board bd,StaticEval stat,ControlPanel cp, Hex7 main) {
      this.bd = bd ;
      this.stat = stat ;
      this.cp = cp ;
      this.main = main ;
   }
   // main routine
   void makeMove() { 
      ply = main.getPly() ; move = null ;
      if (ply<=2) move = openingMove() ;
      if (move==null) {
         // set search depth (ply)
         if (cp.getLevel()==cp.LEVEL1) maxDepth = 4 ; 
         else maxDepth = 6 ;
         // search game tree
         best = abSearch(COMPUTER,PLAYER_WIN,COMPUTER_WIN,0) ;
         move = best.getMove() ;
      }
      // if somehow there's still no move, make a random one
      if (move==null) move = randomMove() ;
      bd.computerMove(move,main.getColor(COMPUTER)) ;
   }
   // returns random integer from 0 to n-1 inclusive
   private int randomInt(int n) { return (int)(n*Math.random()) ; }

   private Point randomMove() {
      int x, y ;

      while (true) {
         if (bd.isEmpty(x=randomInt(bd.SIZE),y=randomInt(bd.SIZE))) break ;
      }
      return new Point(x,y) ;
   }
   private Point openingMove() {
      if ((ply==2)&&(cp.getLevel()==cp.LEVEL2)) {
         // search hexes array for response
         for (int i=0; i<hexes.length; i++)
            if (!bd.isEmpty(hexes[i][0],hexes[i][1]))
               return new Point(hexes[i][2],hexes[i][3]) ;         
      }
      // first move is always on inner diamond (except the center!) or 
      // the 2 hexes on main diagonal from middle ring (1-1 and 5-5)
      else if (ply==1) {
         int i = randomInt(10) + 1 ;
         return new Point(hexes[i][0],hexes[i][1]) ;
      }
      return null ;
   }
   // move generator
   private byte[] moveGen(int depth,int color) {
      // uses 3 main arrays:
      //    "list1" contains all possible moves for white & black, 
      //       with values from shallow search
      //    "list2" contains n-best moves, with values
      //    "ml" is final movelist: byte array with n-best moves only

      int temp[]=new int[3], moveListLength=0, list1ptr=0, list2ptr=0,
          list1[][]=new int[100][3], list2[][] ; 
      byte ml[] ;

      // generate list1
      for (int i=0; i<list1.length; i++ ) list1[i][2] = -999999 ;
      for (int i=0; i<hexes.length; i++) {
         // first get value of move for current color
         if (bd.setHex(hexes[i][0],hexes[i][1],color)) {
            list1[list1ptr][0] = hexes[i][0] ;
            list1[list1ptr][1] = hexes[i][1] ;
            list1[list1ptr][2] = stat.eval(color) ;
            bd.setEmpty(hexes[i][0],hexes[i][1]) ;

            if ((depth==0)&&(list1[list1ptr][2]==stat.WIN_VALUE)) {
               // found winning move
               // return it alone NOW
               ml = new byte[1] ;
               ml[0] = (byte)( (list1[list1ptr][0]*10) + list1[list1ptr][1] ) ;
               return ml ;
            }
            else list1ptr++ ;
            // get value of Black move
            bd.setHex(hexes[i][0],hexes[i][1],main.getOtherColor(color)) ;
            list1[list1ptr][0] = hexes[i][0] ;
            list1[list1ptr][1] = hexes[i][1] ;
            list1[list1ptr++][2] = stat.eval(main.getOtherColor(color)) ;
            bd.setEmpty(hexes[i][0],hexes[i][1]) ;
         }
      }
      // set movelist max size
      if (cp.getLevel()==cp.LEVEL1) moveListLength = 16 ;
      else {
         // at Advanced level, n-best is tapered
         if (depth<2) moveListLength = 16 ;
         else moveListLength = 14 ;
      }
      // initialize list2
      list2 = new int[moveListLength][3] ;
      // Shellsort list1 by value
      for (int h=list1.length/2; h>0; h/=2) {
         for (int i=h; i<list1.length; i++) {
            for (int k=0; k<3; k++) temp[k] = list1[i][k] ; 
            int j = i ;
            for ( ; j>=h && temp[2]>list1[j-h][2] ; j-=h)
               for (int k=0; k<3; k++) list1[j][k] = list1[j-h][k] ;

            for (int k=0; k<3; k++) list1[j][k] = temp[k] ;
         }
      }
      // copy top entries to list2
      // if duplicate move, add value of both entries
      list1ptr = list2ptr = 0 ;

      outer:
      while (list2ptr<moveListLength) {
         // quit if no more possible moves listed
         if (list1[list1ptr][2]==-999999) break ;
         for (int i=0; i<list2ptr; i++) {
            if ( (list1[list1ptr][0]==list2[i][0])&&
                 (list1[list1ptr][1]==list2[i][1]) ) {
               // duplicate found
               list2[i][2] += list1[list1ptr++][2] ;
               continue outer ;
            }
         }
         // copy entry to list2
         list2[list2ptr][0] = list1[list1ptr][0] ;
         list2[list2ptr][1] = list1[list1ptr][1] ;
         list2[list2ptr++][2] = list1[list1ptr++][2] ;
      }
      // now we know how big to make ml
      ml = new byte[list2ptr] ;
      // sort list2 by value
      for (int h=list2.length/2; h>0; h/=2) {
         for (int i=h; i<list2.length; i++) {
            for (int k=0; k<3; k++) temp[k] = list2[i][k] ; 
            int j = i ;
            for ( ; j>=h && temp[2]>list2[j-h][2] ; j-=h)
               for (int k=0; k<3; k++) list2[j][k] = list2[j-h][k] ;

            for (int k=0; k<3; k++) list2[j][k] = temp[k] ;
         }
      }
      // copy list2 to ml: 1-D byte array, and return
      for (int i=0; i<ml.length; i++) 
         ml[i] = (byte)( (list2[i][0]*10) + list2[i][1] ) ;
      return ml ;
   } 
   // negamax game tree search, with alpha-beta
   private BestMove abSearch(int side,int alpha,int beta,int depth) {
      int otherSide, baseCase, value, x, y ;
      int color = main.getColor(side) ;
      BestMove reply ;
      Point saveMove=null ;
      
      // base to end recursion
      baseCase = stat.eval(main.getColor(COMPUTER)) ;
      if ((depth==maxDepth)||(baseCase==-100000)||(baseCase==100000)) 
         return new BestMove(baseCase) ;
      // cutoff for valuable computer move
      if ((depth==maxDepth-1)&&(baseCase>100)) 
         return new BestMove(baseCase) ;
      
      // set initial values
      if (side==COMPUTER) { otherSide = PLAYER ; value = alpha ; }
      else { otherSide = COMPUTER ; value = beta ; }
      // loop through move list
      byte ml[] = moveGen(depth,color) ;
      for (int i=0; i<ml.length; i++) {
         x = (int)(ml[i]/10) ; y = ml[i] % 10 ;
         if (bd.setHex(x,y,color)) {
            reply = abSearch(otherSide,alpha,beta,depth+1) ;
            bd.setEmpty(x,y) ;
         
            if ( ((side==COMPUTER)&&(reply.getValue()>value))||
                 ((side==PLAYER)&&(reply.getValue()<value)) ) {
               // save better move 
               if (side==COMPUTER) alpha = value = reply.getValue() ;
               else beta = value = reply.getValue() ;
               saveMove = new Point(x,y) ;
               // refutation
               if (alpha >= beta) break ;
            }
         }
      }
      return new BestMove(value,saveMove) ;
   }
}
/******** BEST MOVE ********
   temporary storage of candidate moves
*/   
final class BestMove {
   private Point move ;
   private int value ;
   
   // constructors
   BestMove() { }
   BestMove(int v) { this(v, null) ; }
   BestMove(int v, Point p) {
      value = v ;
      move = p ;
   }
   Point getMove() { return move ; }
   int getValue() { return value ; }
}
/******** STATIC EVAL ********
   static evaluator function - produces an integer value for
   any given board position for either side
*/   
final class StaticEval {
   static final int SIZE=7, EMPTY=0, WHITE=1, BLACK=2, WIN_VALUE=100000,
                    SW=1, SE=2, E=3, NE=4, NW=5, W=6, 
                    SW_B=7, S_B=8, SE_B=9, NE_B=10, N_B=11, NW_B=12 ;
   // board areas
   static final int[][] WhiteAreas = { {3,4,4,4,4,4,3},
                                       {3,2,2,2,2,2,3},
                                       {3,2,1,1,1,2,3},
                                       {3,2,1,0,1,2,3},
                                       {3,2,1,1,1,2,3},
                                       {3,2,2,2,2,2,3},
                                       {3,4,4,4,4,4,3} },
                        BlackAreas = { {3,3,3,3,3,3,3},
                                       {4,2,2,2,2,2,4},
                                       {4,2,1,1,1,2,4},
                                       {4,2,1,0,1,2,4},
                                       {4,2,1,1,1,2,4},
                                       {4,2,2,2,2,2,4},
                                       {3,3,3,3,3,3,3} } ;
   private boolean isFullChainWhite, isFullChainBlack ;
   private int stack[]=new int[25], stackPtr=0 ;
   // weights 
   private int A0x, A1x, A2x, A3x, A4x, B1x, B2x, L0x, L1x, E1x, E2x, C1x, C2x, C4x, C5x, 
               chainBaseX, chainPctX ;
   // parametric function values
   private int wA0, wA1, wA2, wA3, wA4, bA0, bA1, bA2, bA3, bA4,
               w2Bridge, b2Bridge, w1Bridge, b1Bridge, w1Link, b1Link, w0Link, b0Link,
               w2Edge, w1Edge, b2Edge, b1Edge,
               wChainLink, wChainBridge, wLinkPct, bChainLink, bChainBridge, bLinkPct,
               w5Chain, w4Chain, w2Chain, w1Chain, b5Chain, b4Chain, b2Chain, b1Chain,
               wVal, bVal ;
   private Board bd ;
   private Hex7 main ;

   // constructor
   StaticEval(Board bd,Hex7 main) { 
      this.bd = bd; 
      this.main = main ;
   }
   // initializes weights and function values
   private void initValues() {
      A0x = 8 ; A1x = 4 ; A2x = 2 ; A3x = 1 ; A4x = 0 ;
      B2x = 8 ; B1x = 4 ; L1x = 3 ; L0x = 0 ; E2x = 4 ; E1x = 2 ;
      C5x = 500 ; C4x = 240 ; C2x = 175 ; C1x = 110 ;

      chainBaseX = 1000 ; chainPctX = 90 ;
      isFullChainWhite = isFullChainBlack = false ;
      wLinkPct = bLinkPct = 0 ;

      wA0 = wA1 = wA2 = wA3 = wA4 = 0 ; 
      bA0 = bA1 = bA2 = bA3 = bA4 = 0 ; 
      w2Bridge = b2Bridge = w1Bridge = b1Bridge = 0 ;
      w1Link = b1Link = w0Link = b0Link = 0 ;
      w2Edge = b2Edge = w1Edge = b1Edge = 0 ;
      wChainLink = wChainBridge = 0 ;
      bChainLink = bChainBridge = 0 ;
      w5Chain = w4Chain = w2Chain = w1Chain = 0 ;
      b5Chain = b4Chain = b2Chain = b1Chain = 0 ;
   }
   // main evaluation method
   int eval(int evalColor) { 

      // first see if it's all over
      if (bd.isWin(evalColor)) return WIN_VALUE ;
      else if (bd.isWin(main.getOtherColor(evalColor))) return -WIN_VALUE ;
      // no win, start eval
      initValues() ;
      // calculate value of chains
      calcChains() ;
      // calculate value for each hex
      calcHex() ;
      // now add everything up
      wVal = (A0x*wA0)+(A1x*wA1)+(A2x*wA2)+(A3x*wA3)+(A4x*wA4) ;
      wVal += (B2x*w2Bridge)+(B1x*w1Bridge)+(L1x*w1Link)+(L0x*w0Link)+
              (E2x*w2Edge)+(E1x*w1Edge) ;
      wVal += (C5x*w5Chain)+(C4x*w4Chain)+(C2x*w2Chain)+(C1x*w1Chain) ;
      if (isFullChainWhite) wVal += chainBaseX + (chainPctX * wLinkPct) ; 

      bVal = (A0x*bA0)+(A1x*bA1)+(A2x*bA2)+(A3x*bA3)+(A4x*bA4) ;
      bVal += (B2x*b2Bridge)+(B1x*b1Bridge)+(L1x*b1Link)+(L0x*b0Link)+
              (E2x*b2Edge)+(E1x*b1Edge) ;
      bVal += (C5x*b5Chain)+(C4x*b4Chain)+(C2x*b2Chain)+(C1x*b1Chain) ;
      if (isFullChainBlack) bVal += chainBaseX + (chainPctX * bLinkPct) ; 
      // and send it back, negamax-style
      return (evalColor==WHITE)? 
         (int)(wVal - bVal) : (int)(bVal - wVal) ;
   }
   // look for chains
   private void calcChains() {

      bd.clearVisited() ;
      // White
      for (int j=0; j<4; j++) {
         for (int i=0; i<SIZE; i++) {
         
            if ((bd.isWhite(i,j))&&(!bd.isVisited(i,j))) {
               stackPtr = wChainLink = wChainBridge = 0 ;
               // first check for valid start
               if (isOpen(i,j,NE)) {
                  bd.setVisited(i,j,true) ;
                  switch (searchChainWhite(i,j)) {
                     // full chain
                     case 8: if (j==0) wChainLink++ ;
                             else wChainBridge++ ;
                             wLinkPct = (int)( 100*(((float)wChainLink/(wChainLink + wChainBridge))) ) ;
                             isFullChainWhite = true ;
                             return ;
                     case 5: w5Chain++ ; break ;
                     case 4: w4Chain++ ; break ;
                     case 2: w2Chain++ ; break ;
                     case 1: if ( ((i==4)||(i==5))&&(j==3)&&(bd.isWhite(4,4)) ) 
                                w1Chain++ ;
                             else if ( (i==2)&&(j==2)&&((bd.isWhite(2,3))||(bd.isWhite(1,3))) ) 
                                w1Chain++ ;
                  }
               }
            }
         }
      }
      // Black
      for (int i=0; i<4; i++) {
         for (int j=0; j<SIZE; j++) {
         
            if ((bd.isBlack(i,j))&&(!bd.isVisited(i,j))) {
               stackPtr = bChainLink = bChainBridge = 0 ;
               // first check for valid start
               if (isOpen(i,j,W)) {
                  bd.setVisited(i,j,true) ;
                  switch (searchChainBlack(i,j)) {
                     // full chain
                     case 8: if (i==0) bChainLink++ ;
                             else bChainBridge++ ;
                             bLinkPct = (int)( 100*(((float)bChainLink/(bChainLink + bChainBridge))) ) ;
                             isFullChainBlack = true ;
                             return ;
                     case 5: b5Chain++ ; break ;
                     case 4: b4Chain++ ; break ;
                     case 2: b2Chain++ ; break ;
                     case 1: if ( ((j==4)||(j==5))&&(i==3)&&(bd.isBlack(4,4)) ) 
                                b1Chain++ ;
                             else if ( (i==2)&&(j==2)&&((bd.isBlack(3,2))||(bd.isBlack(3,1))) ) 
                                b1Chain++ ;
                  }
               }
            }
         }
      }
   }
   // DFS for chains - returns "absolute" length of chain 
   //    (number of rows or columns it spans including edges)
   private int searchChainWhite(int x,int y) {
      int direction=0, currLength=0, maxLength=0, saveLength=0, localMaxLength=0 ;

      // initialize length counters
      if (y==0) currLength = maxLength = 1 ;
      else if (y==1) currLength = maxLength = 2 ;

      while (true) {
         // look for a link
         direction = bd.scanLinks(x,y,WHITE) ;
         // nope, look for a bridge
         if (direction==0) direction = bd.scanBridges(x,y,WHITE) ;
         if (direction!=0) {
            // found something - bump the counter
            if (direction>6) wChainBridge++ ;
            else wChainLink++ ;
            // make hex current
            stack[stackPtr++] = direction ;
            switch (direction) {
               case SW: y++ ; currLength++ ; break ;
               case SE: x++ ; y++ ; currLength++ ; break ;
               case E : x++ ; break ;
               case NE: y-- ; currLength-- ; break ;
               case NW: x-- ; y-- ; currLength-- ; break ;
               case W : x-- ; break ;
               case SW_B: x--  ; y++  ; currLength++ ; break ;
               case S_B : x++  ; y+=2 ; currLength+=2 ; break ;
               case SE_B: x+=2 ; y++  ; currLength++ ; break ;
               case NE_B: x++  ; y--  ; currLength-- ; break ;
               case N_B : x--  ; y-=2 ; currLength-=2 ; break ;
               case NW_B: x-=2 ; y--  ; currLength-- ; break ;
            }
            // check overall length
            if (currLength>maxLength) maxLength = currLength ;
            // check for local max if we're not moving forward
            else if (localMaxLength<maxLength)
               if ( (y>=3)&&(maxLength>0)&&(isOpen(x,y,SW)) ) localMaxLength = maxLength ;
            // check if we've reached the bottom
            if (y==(SIZE-1)) { wChainLink++ ; return maxLength+1 ; }
            if (y==(SIZE-2)) if (isOpen(x,y,SW)) { wChainBridge++ ; return maxLength+2 ; }
            // nope, look for next link/bridge
            bd.setVisited(x,y,true) ;
         }
         else {
            if (stackPtr==0)
               // all chains visited, return the longest we've found
               return (saveLength>localMaxLength)? saveLength : localMaxLength ;
            // at end of path, check chain and back up
            if (saveLength<currLength)
               if ( (y>=3)&&(currLength>0)&&(isOpen(x,y,SW)) ) saveLength = currLength ;
            direction = stack[--stackPtr] ;
            // decrement the counter
            if (direction>6) wChainBridge-- ;
            else wChainLink-- ;

            switch (direction) {
               case SW: y-- ; currLength-- ; maxLength-- ; break ;
               case SE: x-- ; y-- ; currLength-- ; maxLength-- ; break ;
               case E : x-- ; break ;
               case NE: y++ ; currLength++ ; break ;
               case NW: x++ ; y++ ; currLength++ ; break ;
               case W : x++ ; break ;
               case SW_B: x++  ; y--  ; currLength-- ; maxLength-- ; break ;
               case S_B : x--  ; y-=2 ; currLength-=2 ; maxLength-=2 ; break ;
               case SE_B: x-=2 ; y--  ; currLength-- ; maxLength-- ; break ;
               case NE_B: x--  ; y++  ; currLength++ ; break ;
               case N_B : x++  ; y+=2 ; currLength+=2 ; break ;
               case NW_B: x+=2 ; y++  ; currLength++ ; break ;
            }
         }
      }
   }
   private int searchChainBlack(int x,int y) {
      int direction=0, currLength=0, maxLength=0, saveLength=0, localMaxLength=0 ;

      // initialize length counters
      if (x==0) currLength = maxLength = 1 ;
      else if (x==1) currLength = maxLength = 2 ;

      while (true) {
         // look for a link
         direction = bd.scanLinks(x,y,BLACK) ;
         // nope, look for a bridge
         if (direction==0) direction = bd.scanBridges(x,y,BLACK) ;
         if (direction!=0) {
            // found something - bump the counter
            if (direction>6) bChainBridge++ ;
            else bChainLink++ ;
            // make hex current
            stack[stackPtr++] = direction ;
            switch (direction) {
               case SW: y++ ; break ;
               case SE: x++ ; y++ ; currLength++ ; break ;
               case E : x++ ; currLength++ ; break ;
               case NE: y-- ; break ;
               case NW: x-- ; y-- ; currLength-- ; break ;
               case W : x-- ; currLength-- ; break ;
               case SW_B: x--  ; y++  ; currLength-- ; break ;
               case S_B : x++  ; y+=2 ; currLength++ ; break ;
               case SE_B: x+=2 ; y++  ; currLength+=2 ; break ;
               case NE_B: x++  ; y--  ; currLength++ ; break ;
               case N_B : x--  ; y-=2 ; currLength-- ; break ;
               case NW_B: x-=2 ; y--  ; currLength-=2 ; break ;
            }
            // check overall length
            if (currLength>maxLength) maxLength = currLength ;
            // check for local max if we're not moving forward
            else if (localMaxLength<maxLength)
               if ( (x>=3)&&(maxLength>0)&&(isOpen(x,y,E)) ) localMaxLength = maxLength ;
            // check if we've reached the bottom
            if (x==(SIZE-1)) { bChainLink++ ; return maxLength+1 ; }
            if (x==(SIZE-2)) if (isOpen(x,y,E)) { bChainBridge++ ; return maxLength+2 ; }
            // nope, look for next link/bridge
            bd.setVisited(x,y,true) ;
         }
         else {
            if (stackPtr==0)
               // all chains visited, return the longest we've found
               return (saveLength>localMaxLength)? saveLength : localMaxLength ;
            // at end of path, check chain and back up
            if (saveLength<currLength)
               if ( (x>=3)&&(currLength>0)&&(isOpen(x,y,E)) ) saveLength = currLength ;
            direction = stack[--stackPtr] ;
            // decrement the counter
            if (direction>6) bChainBridge-- ;
            else bChainLink-- ;

            switch (direction) {
               case SW: y-- ; break ;
               case SE: x-- ; y-- ; currLength-- ; maxLength-- ; break ;
               case E : x-- ; currLength-- ; maxLength-- ; break ;
               case NE: y++ ; break ;
               case NW: x++ ; y++ ; currLength++ ; break ;
               case W : x++ ; currLength++ ; break ;
               case SW_B: x++  ; y--  ; currLength++ ; break ;
               case S_B : x--  ; y-=2 ; currLength-- ; maxLength-- ; break ;
               case SE_B: x-=2 ; y--  ; currLength-=2 ; maxLength-=2 ; break ;
               case NE_B: x--  ; y++  ; currLength-- ;  maxLength-- ; break ;
               case N_B : x++  ; y+=2 ; currLength++ ; break ;
               case NW_B: x+=2 ; y++  ; currLength+=2 ; break ;
            }
         }
      }
   }
   // checks for open edge connection from x,y in given direction
   private boolean isOpen(int x,int y,int dir) {

      switch (dir) {
         case NE: 
            switch (y) {
               case 0: return true ;
               case 1: 
                  // for edge bridges the hexes must be empty
                  if ( (x!=0)&&(bd.isEmpty(x-1,0))&&(bd.isEmpty(x,0)) ) return true ;
                  break ;
               case 2: 
                  if ( (x>1)&&(!bd.isBlack(x-1,1))&&(!bd.isBlack(x,1))&&
                       (!bd.isBlack(x-2,0))&&(!bd.isBlack(x-1,0))&&(!bd.isBlack(x,0)) ) {
                     // above is good, now check one side or the other
                     switch (x) {
                        case 2: 
                           if ( (!bd.isBlack(3,2))&&(!bd.isBlack(3,1))&&(!bd.isBlack(3,0)) )
                              return true ;
                           break ;
                        case 6: 
                           if ( (!bd.isBlack(5,2))&&(!bd.isBlack(4,1))&&(!bd.isBlack(3,0)) )
                              return true ;
                           break ;
                        default: 
                           if ( ((!bd.isBlack(x+1,2))&&(!bd.isBlack(x+1,1))&&(!bd.isBlack(x+1,0)))||
                                ((!bd.isBlack(x-1,2))&&(!bd.isBlack(x-2,1))&&(!bd.isBlack(x-3,0))) )
                              return true ;
                           break ;
                     }
                  }
                  break ;
               case 3: 
                  if ( ((x==4)&&(!bd.isBlack(5,3)))||
                       ((x==5)&&(!bd.isBlack(4,3))&&(!bd.isBlack(6,3))) ) {
                     for (int i=2; i<7; i++) if (bd.isBlack(i,2)) return false ;
                     for (int i=1; i<7; i++) if (bd.isBlack(i,1)) return false ;
                     for (int i=0; i<7; i++) if (bd.isBlack(i,0)) return false ;
                     return true ;
                  }
                  break ;
            }
            break ;

         case SW: 
            switch (y) {
               case 3: 
                  if ( ((x==2)&&(!bd.isBlack(1,3)))||
                       ((x==1)&&(!bd.isBlack(0,3))&&(!bd.isBlack(2,3))) ) {
                     for (int i=0; i<5; i++) if (bd.isBlack(i,4)) return false ;
                     for (int i=0; i<6; i++) if (bd.isBlack(i,5)) return false ;
                     for (int i=0; i<7; i++) if (bd.isBlack(i,6)) return false ;
                     return true ;
                  }
                  break ;
               case 4: 
                  if ( (x<5)&&(!bd.isBlack(x,5))&&(!bd.isBlack(x+1,5))&&
                       (!bd.isBlack(x,6))&&(!bd.isBlack(x+1,6))&&(!bd.isBlack(x+2,6)) ) {
                     // below is good, now check one side or the other
                     switch (x) {
                        case 0: 
                           if ( (!bd.isBlack(1,4))&&(!bd.isBlack(2,5))&&(!bd.isBlack(3,6)) )
                              return true ;
                           break ;
                        case 4: 
                           if ( (!bd.isBlack(3,4))&&(!bd.isBlack(3,5))&&(!bd.isBlack(3,6)) )
                              return true ;
                           break ;
                        default: 
                           if ( ((!bd.isBlack(x-1,4))&&(!bd.isBlack(x-1,5))&&(!bd.isBlack(x-1,6)))||
                                ((!bd.isBlack(x+1,4))&&(!bd.isBlack(x+2,5))&&(!bd.isBlack(x+3,6))) )
                              return true ;
                           break ;
                     }
                  }
                  break ;
               // for edge bridges the hexes must be empty
               case 5: if ( (x!=6)&&(bd.isEmpty(x,6))&&(bd.isEmpty(x+1,6)) ) return true ;
                       break ;
               case 6: return true ;
            }
            break ;

         case W: 
            switch (x) {
               case 0: return true ;
               case 1: 
                  // for edge bridges the hexes must be empty
                  if ( (y!=0)&&(bd.isEmpty(0,y-1))&&(bd.isEmpty(0,y)) ) return true ;
                  break ;
               case 2: 
                  if ( (y>1)&&(!bd.isWhite(1,y-1))&&(!bd.isWhite(1,y))&&
                       (!bd.isWhite(0,y-2))&&(!bd.isWhite(0,y-1))&&(!bd.isWhite(0,y)) ) {
                     // left is good, now check one side or the other
                     switch (y) {
                        case 2: 
                           if ( (!bd.isWhite(2,3))&&(!bd.isWhite(1,3))&&(!bd.isWhite(0,3)) )
                              return true ;
                           break ;
                        case 6: 
                           if ( (!bd.isWhite(2,5))&&(!bd.isWhite(1,4))&&(!bd.isWhite(0,3)) )
                              return true ;
                           break ;
                        default: 
                           if ( ((!bd.isWhite(2,y+1))&&(!bd.isWhite(1,y+1))&&(!bd.isWhite(0,y+1)))||
                                ((!bd.isWhite(2,y-1))&&(!bd.isWhite(1,y-2))&&(!bd.isWhite(0,y-3))) )
                              return true ;
                           break ;
                     }
                  }
                  break ;
               case 3: 
                  if ( ((y==4)&&(!bd.isWhite(3,5)))||
                       ((y==5)&&(!bd.isWhite(3,4))&&(!bd.isWhite(3,6))) ) {
                     for (int j=2; j<7; j++) if (bd.isWhite(2,j)) return false ;
                     for (int j=1; j<7; j++) if (bd.isWhite(1,j)) return false ;
                     for (int j=0; j<7; j++) if (bd.isWhite(0,j)) return false ;
                     return true ;
                  }
                  break ;
            }
            break ;

         case E: 
            switch (x) {
               case 3: 
                  if ( ((y==2)&&(!bd.isWhite(3,1)))||
                       ((y==1)&&(!bd.isWhite(3,0))&&(!bd.isWhite(3,2))) ) {
                     for (int j=0; j<5; j++) if (bd.isWhite(4,j)) return false ;
                     for (int j=0; j<6; j++) if (bd.isWhite(5,j)) return false ;
                     for (int j=0; j<7; j++) if (bd.isWhite(6,j)) return false ;
                     return true ;
                  }
                  break ;
               case 4: 
                  if ( (y<5)&&(!bd.isWhite(5,y))&&(!bd.isWhite(5,y+1))&&
                       (!bd.isWhite(6,y))&&(!bd.isWhite(6,y+1))&&(!bd.isWhite(6,y+2)) ) {
                     // right is good, now check one side or the other
                     switch (y) {
                        case 0: 
                           if ( (!bd.isWhite(4,1))&&(!bd.isWhite(5,2))&&(!bd.isWhite(6,3)) )
                              return true ;
                           break ;
                        case 4: 
                           if ( (!bd.isWhite(4,3))&&(!bd.isWhite(5,3))&&(!bd.isWhite(6,3)) )
                              return true ;
                           break ;
                        default: 
                           if ( ((!bd.isWhite(4,y-1))&&(!bd.isWhite(5,y-1))&&(!bd.isWhite(6,y-1)))||
                                ((!bd.isWhite(4,y+1))&&(!bd.isWhite(5,y+2))&&(!bd.isWhite(6,y+3))) )
                              return true ;
                           break ;
                     }
                  }
                  break ;
               // for edge bridges the hexes must be empty
               case 5: if ( (y!=6)&&(bd.isEmpty(6,y))&&(bd.isEmpty(6,y+1)) ) return true ;
                       break ;
               case 6: return true ;
            }
            break ;
      }
      return false ;
   }
   // calculates value for each piece
   //    always calcs area value
   //    calcs bridges & links only if no chains
   private void calcHex() {

      for (int i=0; i<SIZE; i++)
         for (int j=0; j<SIZE; j++)
            if (!bd.isEmpty(i,j)) {

               if (bd.getHex(i,j)==WHITE) {
                  switch (WhiteAreas[i][j]) {
                     case 0: wA0++ ; break ;
                     case 1: wA1++ ; break ;
                     case 2: wA2++ ; break ;
                     case 3: wA3++ ; break ;
                     case 4: wA4++ ; break ;
                  }
                  if (!isChain(WHITE)) calcBridgesLinks(i,j,WHITE) ;
               }
               else {
                  switch (BlackAreas[i][j]) {
                     case 0: bA0++ ; break ;
                     case 1: bA1++ ; break ;
                     case 2: bA2++ ; break ;
                     case 3: bA3++ ; break ;
                     case 4: bA4++ ; break ;
                  }
                  if (!isChain(BLACK)) calcBridgesLinks(i,j,BLACK) ;
               }
            }
   }
   // returns true if any type of chain exists
   private boolean isChain(int color) {
      if (color==WHITE) {
         if ( (isFullChainWhite)||(w5Chain!=0)||(w4Chain!=0)||(w2Chain!=0)||(w1Chain!=0) )
            return true ;
      }
      else {
         if ( (isFullChainBlack)||(b5Chain!=0)||(b4Chain!=0)||(b2Chain!=0)||(b1Chain!=0) )
            return true ;
      }
      return false ;
   }
   // calculates bridges and links for one hex
   private void calcBridgesLinks(int i,int j,int color) {

      calcEdges(i,j,color) ;

      if (calc2Bridge(i,j,color)) {
         if (color==WHITE) {
            calc1BridgeRight(i,j,color) ;
            if (!calc1BridgeLeft(i,j,color)) calc0Link(i,j,color) ;
         }
         else {
            calc1BridgeLeft(i,j,color) ;
            if (!calc1BridgeRight(i,j,color)) calc0Link(i,j,color) ;
         }
      }
      else {
         if (color==WHITE) {
            if (!calc1BridgeRight(i,j,color)) calc1LinkRight(i,j,color) ;
            if (!calc1BridgeLeft(i,j,color)) calc1LinkLeft(i,j,color) ;
            calc0Link(i,j,color) ;
         }
         else {
            if (!calc1BridgeLeft(i,j,color)) calc1LinkLeft(i,j,color) ;
            if (!calc1BridgeRight(i,j,color)) calc1LinkRight(i,j,color) ;
            calc0Link(i,j,color) ;
         }
      }
   }
   // calculates edge connections for one hex
   private void calcEdges(int i,int j,int color) {
      if (color==WHITE) {
         if ( ((j==0)&&(i!=(SIZE-1))) || ((j==(SIZE-1))&&(i!=0)) ) w1Edge++ ;
         else if ( (j==1)&&(isOpen(i,j,NE)) ) w2Edge++ ;
         else if ( (j==(SIZE-2))&&(isOpen(i,j,SW)) ) w2Edge++ ;
      } else {
         if ( ((i==0)&&(j!=(SIZE-1))) || ((i==(SIZE-1))&&(j!=0)) ) b1Edge++ ;
         else if ( (i==1)&&(isOpen(i,j,W)) ) b2Edge++ ;
         else if ( (i==(SIZE-2))&&(isOpen(i,j,E)) ) b2Edge++ ;
      }
   }
   private boolean calc2Bridge(int i,int j,int color) {
      if (color==WHITE) {
         if (bd.isBridgeS(i,j,WHITE)) { w2Bridge++ ; return true ; }
      } 
      else if (bd.isBridgeSE(i,j,BLACK)) { b2Bridge++ ; return true ; }
      return false ;
   }
   private boolean calc1BridgeLeft(int i,int j,int color) {
      if (color==WHITE) {
         if (bd.isBridgeSE(i,j,WHITE)) { w1Bridge++ ; return true ; }
      } 
      else if (bd.isBridgeNE(i,j,BLACK)) { b1Bridge++ ; return true ; }
      return false ;
   }
   private boolean calc1BridgeRight(int i,int j,int color) {
      if (color==WHITE) {
         if (bd.isBridgeSW(i,j,WHITE)) { w1Bridge++ ; return true ; }
      } 
      else if (bd.isBridgeS(i,j,BLACK)) { b1Bridge++ ; return true ; }
      return false ;
   }
   private boolean calc1LinkLeft(int i,int j,int color) {
      if (color==WHITE) {
         if (bd.isLinkSE(i,j,WHITE)) { w1Link++ ; return true ; }
      }
      else if (bd.isLinkE(i,j,BLACK)) { b1Link++ ; return true ; }
      return false ;
   }
   private boolean calc1LinkRight(int i,int j,int color) {
      if (color==WHITE) { 
         if (bd.isLinkSW(i,j,WHITE)) { w1Link++ ; return true ; } 
      }
      else if (bd.isLinkSE(i,j,BLACK)) { b1Link++ ; return true ; }
      return false ;
   }
   private void calc0Link(int i,int j,int color) {
      if (color==WHITE) { if (bd.isLinkE(i,j,WHITE)) w0Link++ ; } 
      else if (bd.isLinkSW(i,j,BLACK)) b0Link++ ;
   }
}
/******** BOARD CANVAS ********
   draws board, handles conversion of hexes/pixels
*/   
final class BoardCanvas extends Canvas {
   // x-y coordinates for Giflet images
   final static int whiteStoneCoord[]={0,0,20,20}, blackStoneCoord[]={21,0,20,20} ;
   private Image bufferImage, boardImage, gifletImage, whiteStoneImage, blackStoneImage ;
   private MediaTracker tracker ;
   private Graphics buffer ;
   private Hex7 main ;
   
   // constructor
   BoardCanvas(Image bI,Image gI,Hex7 main) {
      boardImage = bI ;
      gifletImage = gI ;
      this.main = main ;

      // extract images from giflet
      tracker = new MediaTracker(main) ;
      whiteStoneImage = extractImage(whiteStoneCoord) ;
      blackStoneImage = extractImage(blackStoneCoord) ;

      try { tracker.waitForAll() ; }
      catch (InterruptedException e) {}
   }
	void drawBoard(Board bd) {
      if (buffer==null) {
         bufferImage = createImage(main.bcWidth,main.bcHeight) ;
         buffer = bufferImage.getGraphics() ;
      }
		// draw the board
		buffer.drawImage(boardImage,0,0,this) ;
		// add the pieces
		for (int i = 0; i < bd.SIZE; i++) {
			for (int j = 0; j < bd.SIZE; j++) {
				if (!bd.isEmpty(i,j)) {
					buffer.drawImage(bd.getHex(i,j)==1?whiteStoneImage:blackStoneImage,
						hexToPixelX(i,j),hexToPixelY(i,j),this) ;
				}
			}
		}
		repaint() ;
	}
	public void paint(Graphics g) { update(g) ; }
	public void update(Graphics g) { 
      if (bufferImage!=null) g.drawImage(bufferImage,0,0,this) ;
   }
	
	Point pixelToHex(int x,int y) {
      int I=-1, J=-1, i=-1, j=-1 ;
      if ( (x>50)&&(x<395)&&(y>22)&&(y<223) ) {
         if (((x-51)%27)<20) {
            I = (int)((x-51)/27) ;
            if (I%2==0) { if (((y-23)%30)<20) J = ((int)((y-23)/30))*2 ; } 
            else if (((y-38)%30)<20) { J = (((int)((y-38)/30))*2)+1 ; }
            if (J>=0) {
               i = (int)((I+J-6)/2) ;
               j = (int)((J-I+6)/2) ;
               if ( (i>=0)&&(i<7)&&(j>=0)&&(j<7) ) return new Point(i,j) ;
            }
         }
      }            
      return new Point(-1,-1) ;
	}
	int hexToPixelX(int i,int j) { return ((i-j)*27)+213 ; }
	int hexToPixelY(int i,int j) { return ((i+j)*15)+23 ; }

	// mouse press event handler - passes event back to main class
	public boolean mouseDown(Event e,int x,int y) {	
      Point p = pixelToHex(x,y) ; 
      main.selectHex(p.x,p.y) ;
      return true ; 
   }
   // extracts GIF images
   private Image extractImage(int[] xyCoord) {
      Image newImage ;
      ImageFilter filter;
      ImageProducer producer;

      filter = new CropImageFilter(xyCoord[0],xyCoord[1],xyCoord[2],xyCoord[3]);
      producer = new FilteredImageSource(gifletImage.getSource(), filter);
      newImage = main.createImage(producer);
      tracker.addImage(newImage,0);
      return newImage ;
   }
}
/******** CONTROL PANEL ********
   the UI components
*/   
final class ControlPanel extends Panel {
   static final int LEVEL1=1, LEVEL2=2 ;
   static final Font titleFont=new Font("SansSerif", Font.BOLD, 14),
                     textFont=new Font("SansSerif", Font.PLAIN, 12),
                     statusFont=new Font("Dialog", Font.PLAIN, 12) ;
   private Label lTitle ;
   private Button bNewGame, bMove ;
   private CheckboxGroup cbgSide, cbgLevel ;
   private Checkbox cbWhite, cbBlack, cbBasic, cbAdvanced ;
   private TextField tfStatus ;
   private Hex7 main ;
   
   // constructor
   ControlPanel(Hex7 main) {
      this.main = main ;
      // main layout
      setLayout(new GridLayout(0,1,0,2)) ;
      setBackground(Color.lightGray) ;
      setFont(textFont) ;
      // Title
      add(lTitle=new Label(main.sTITLE,Label.CENTER)) ;
      lTitle.setFont(titleFont) ;
      add(new Label(" ")) ;
      // New Game button
      add(bNewGame=new Button(main.sNEW_GAME)) ;
      // Player Side checkbox group
      cbgSide=new CheckboxGroup() ;
      add(cbWhite=new Checkbox(main.sWHITE,cbgSide,true)) ;
      add(cbBlack=new Checkbox(main.sBLACK,cbgSide,false)) ;
      // Level checkbox group
      add(new Label(main.sLEVEL,Label.CENTER)) ;
      cbgLevel=new CheckboxGroup() ;
      add(cbBasic=new Checkbox(main.sBASIC,cbgLevel,true)) ;
      add(cbAdvanced=new Checkbox(main.sADVANCED,cbgLevel,false)) ;
      // Status textfield
      tfStatus = new TextField(10) ;
      tfStatus.setFont(statusFont) ;
      tfStatus.setForeground(Color.white) ;
      tfStatus.setBackground(Color.black) ;
      tfStatus.setEditable(false) ;
      add(tfStatus) ;      
      validate() ;
   }
   // handle control panel events
   public boolean action(Event e,Object o) { 
      if (main.sNEW_GAME.equals(o)) main.newGame() ; 
      else if (e.target==cbWhite) main.setPlayerColor(main.WHITE) ;
      else if (e.target==cbBlack) main.setPlayerColor(main.BLACK) ;
      else main.setLevel() ;
      return true ;
   }
   int getLevel() {
      if (cbBasic.getState()) return LEVEL1 ;
      else if (cbAdvanced.getState()) return LEVEL2 ;
      return 0 ;
   }
   void setStatus(String s) { tfStatus.setText(s) ; }
}